@@ -60,61 +60,132 @@ Note: the state vector 'mt' returned by pymt64.init has 313 elements instead of
 
 
 
-
-
-double gammln(double xx)
+/* log-gamma function to support some of these distributions. The 
+ * algorithm comes from SPECFUN by Shanjie Zhang and Jianming Jin and their
+ * book "Computation of Special Functions", 1996, John Wiley & Sons, Inc.
+ */
+double loggam(double x)
 {
-	double x,y,tmp,ser;
-	static double cof[6]={76.18009172947146,-86.50532032941677,
-		24.01409824083091,-1.231739572450155,
-		0.1208650973866179e-2,-0.5395239384953e-5};
-	int j;
-
-	y=x=xx;
-	tmp=x+5.5;
-	tmp -= (x+0.5)*log(tmp);
-	ser=1.000000000190015;
-	for (j=0;j<=5;j++) ser += cof[j]/++y;
-	return -tmp+log(2.5066282746310005*ser/x);
+    double x0, x2, xp, gl, gl0;
+    long k, n;
+    
+    double a[10] = {8.333333333333333e-02,-2.777777777777778e-03,
+         7.936507936507937e-04,-5.952380952380952e-04,
+         8.417508417508418e-04,-1.917526917526918e-03,
+         6.410256410256410e-03,-2.955065359477124e-02,
+         1.796443723688307e-01,-1.39243221690590e+00};
+    x0 = x;
+    n = 0;
+    if ((x == 1.0) || (x == 2.0))
+    {
+        return 0.0;
+    }
+    else if (x <= 7.0)
+    {
+        n = (long)(7 - x);
+        x0 = x + n;
+    }
+    x2 = 1.0/(x0*x0);
+    xp = 2*PI;
+    gl0 = a[9];
+    for (k=8; k>=0; k--)
+    {
+        gl0 *= x2;
+        gl0 += a[k];
+    }
+    gl = gl0/x0 + 0.5*log(xp) + (x0-0.5)*log(x0) - x0;
+    if (x <= 7.0)
+    {
+        for (k=1; k<=n; k++)
+        {
+            gl -= log(x0-1.0);
+            x0 -= 1.0;
+        }
+    }
+    return gl;
 }
 
+/*
+  Adapted from rk_poisson_mult by Robert Kern (robert.kern@gmail.com)
+*/
 
-double poisson(UL64 * mt , int * mti , double xm )
+long rk_poisson_mult(UL64 * mt , int * mti , double lam)
 {
-	double gammln(double xx);
-	static double sq,alxm,g,oldm=(-1.0);
-	double em,t,y;
-
-	if (xm < 12.0) {
-		if (xm != oldm) {
-			oldm=xm;
-			g=exp(-xm);
-		}
-		em = -1;
-		t=1.0;
-		do {
-			++em;
-			t *= genrand64_real1(mt,mti) ;
-		} while (t > g);
-	} else {
-		if (xm != oldm) {
-			oldm=xm;
-			sq=sqrt(2.0*xm);
-			alxm=log(xm);
-			g=xm*alxm-gammln(xm+1.0);
-		}
-		do {
-			do {
-				y=tan(PI*genrand64_real1(mt,mti));
-				em=sq*y+xm;
-			} while (em < 0.0);
-			em=floor(em);
-			t=0.9*(1.0+y*y)*exp(em*alxm-gammln(em+1.0)-g);
-		} while (genrand64_real1(mt,mti)  > t);
+  long X;
+  double prod, U, enlam;
+  enlam = exp(-lam);
+  X = 0;
+  prod = 1.0;
+  while (1)
+    {
+      U =  genrand64_real1(mt,mti) ;  // random number on [0,1]-real-interval
+      prod *= U;
+      if (prod > enlam)
+	{
+	  X += 1;
 	}
-	return em;
+      else
+	{
+	  return X;
+	}
+    }
 }
+/*
+  Adapted from rk_poisson_ptrs by Robert Kern (robert.kern@gmail.com)
+*/
+#define LS2PI 0.91893853320467267
+#define TWELFTH 0.083333333333333333333333
+long rk_poisson_ptrs(UL64 * mt , int * mti ,  double lam)
+{
+  long k;
+  double U, V, slam, loglam, a, b, invalpha, vr, us;
+  slam = sqrt(lam);
+  loglam = log(lam);
+  b = 0.931 + 2.53*slam;
+  a = -0.059 + 0.02483*b;
+  invalpha = 1.1239 + 1.1328/(b-3.4);
+  vr = 0.9277 - 3.6224/(b-2);
+  while (1)
+    {
+      U = genrand64_real1(mt,mti)  - 0.5;
+      V = genrand64_real1(mt,mti)  ;
+      us = 0.5 - fabs(U);
+      k = (long)floor((2*a/us + b)*U + lam + 0.43);
+      if ((us >= 0.07) && (V <= vr))
+	{
+	  return k;
+	}
+      if ((k < 0) ||
+	  ((us < 0.013) && (V > us)))
+	{
+	  continue;
+	}
+      if ((log(V) + log(invalpha) - log(a/(us*us)+b)) <=
+	  (-lam + k*loglam - loggam(k+1)))
+	{
+	  return k;
+	}
+    }
+}
+/*
+  Adapted from rk_poisson by Robert Kern (robert.kern@gmail.com)
+*/
 
+long  poisson(UL64 * mt , int * mti , double lam )
+{
+  if (lam >= 10)
+    {
+      return rk_poisson_ptrs(mt,mti, lam);
+    }
+  else if (lam == 0)
+    {
+      return 0;
+    }
+  else
+    {
+      return rk_poisson_mult(mt,mti, lam);
+    }
+}
 
 void uniformn(UL64 * mt , int * mti, double * x , long n )
 
